/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.gui.tool;

import java.awt.Graphics;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import org.apache.log4j.Logger;

import es.prodevelop.gvsig.mobile.common.GvsigMobileException;
import es.prodevelop.gvsig.mobile.common.GvsigMobileGUIConstants;
import es.prodevelop.gvsig.mobile.gui.mainframe.MainWindow;

/**
 * A button that is aimed to be placed in a tool bar, with the ability to
 * contain several buttons inside. Only one button is shown in the tool bar,
 * the rest are accesible trough a list under a combo box style.
 * <p>A <code>TBComboButton</code> must be placed in a <code>ToolBar</code>
 * and contains:</p>
 * <ul>
 * 		<li>One or more <code>TBButton</code>s </li>
 * 		<li>A <code>TBDropDownButton</code> for accessing the buttons</li>
 * </ul>
 * 
 * @author mmontesinos
 * @see ToolBar
 * @see TBButton
 * @see TBDropDownButton
 */	
public class TBComboButton extends TBButton implements IToolContainer, MouseListener{
	
	/*
	 * Array of buttons available in the drop-down list
	 */
	private TBButton[] buttons = new TBButton[0]; //A default button must exist
	/*
	 * The button with an arrow for showing a drop-down menu
	 */
	private TBDropDownButton dropDown;
	private TBMenu menu = null;
	protected static Logger logger = Logger.getLogger(TBComboButton.class);
	
	protected int index = 0;
	
	/**
	 * Creates an instance of a TBComboButton button by calling to the
	 * superclass constructor given the owner where it should
	 * be placed (which must be a <code>ToolBar</code>), the default button
	 * (which must be a <code>TBButton</code> and the
	 * listener of the events for this new instance
	 * @param _owner Toolbar parent of the TBComboButton (<code>ToolBar</code>)
	 * 				 where the button is to be added
	 * @param defaultTBButton Default button of the combo buttons list. The one
	 * 				 that will be seen at the beginning in the Toolbar.
	 * 				 Must be a previously created <code>TBButton</code>
	 * @param ml Listener of Mouse Events, which must be a
	 * 				 <code>MouseListener</code>
	 */
	public TBComboButton(ToolBar _owner,
			 TBButton defaultTBButton, MouseListener ml) {
		super(_owner, defaultTBButton.getIcon(), defaultTBButton.getSelectedIcon(), defaultTBButton.getHelpText(), ml);
		this.setWidth(GvsigMobileGUIConstants.TB_COMBO_BUTTON_WIDTH);
		this.setHeight(GvsigMobileGUIConstants.DEFAULT_BUTTON_HEIGHT);
		
		// Let's create the default button as the first button
		try {
			addButton(defaultTBButton, 0);
		} catch (GvsigMobileException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		index = 0;
		dropDown = new TBDropDownButton(this, this.getMyListener());
		this.add(dropDown);
		
	}
	
	/**
	 * Adds a button to the combo list. The button must be of type
	 * <code>TBButton</code>
	 * 
	 * @param button <code>TBButton</code> to be added to the ToolBar
	 * @param index number of the position that the new button will have in the
	 * 		  array of buttons that the <code>TBComboButton</code> contains.
	 * 		  It must be an <code>int</code> between 0 and the actual size of
	 * 		  the array.
	 * @throws GvsigMobileException
	 */
	public void addButton (TBButton button, int index) throws GvsigMobileException{
	
		int newLength = 0;
		
		if (index < 0){
			throw new GvsigMobileException(GvsigMobileGUIConstants.ERROR_INDEX_UNDER_0);
		} else if (index > buttons.length){
			throw new GvsigMobileException(GvsigMobileGUIConstants.ERROR_TB_BUTTON_GAPS);
		} 
		
		button.setOwner(this);
		
		newLength = buttons.length + 1;
		TBButton[] newButtons = new TBButton [newLength];
		
		if (buttons.length == 0){ //First button to be added
			buttons = newButtons;
		} else{
			if (index > 0){ //Left-side of the array need to be copied
				System.arraycopy (buttons, 0, newButtons, 0, index);
			}
			if (index < newLength - 1){ //Rigth-side of the array need to be copied
				System.arraycopy (buttons, index, newButtons, index + 1, newLength - index - 1);
			}
			buttons = newButtons;
		}
	
		/*
		 * We have a new buttons array with an empty space for the new
		 * TBButton to be created
		 */
		buttons[index] = button;
		//The owner of a TBComboButton must be always a ToolBar
		this.add(buttons[index]);
		buttons[index].setBounds(0, 0, buttons[index].getWidth(), buttons[index].getHeight());
	
	}
	
	/**
	 * Create a Drop-down menu associated to the combo button with
	 * the buttons added to the <code>TBComboButton</code> as the items list
	 *
	 */
	public void createMenu(){
		ToolBar myToolBar;
		myToolBar = ((ToolBar) owner);
		menu = new TBMenu(this, getX(), getY() + getHeight());
		myToolBar.mainWindow.dropMenu = menu;
		
	}
	
	/**
	 * Removes the menu associated to the combo button
	 *
	 */
	public void removeMenu(){
		ToolBar myToolBar;
		
		menu.close();
		menu = null;
		//Let's remove all buttons from the combo and re-add them
		this.removeAll();
		for (int i = 0; i < buttons.length; i++){
			this.add(buttons[i]);
			buttons[i].setVisible(false);
		}
		//Set selected button properly
		buttons[index].setVisible(true);
		buttons[index].setBounds(0, 0, GvsigMobileGUIConstants.DEFAULT_BUTTON_WIDTH, GvsigMobileGUIConstants.DEFAULT_BUTTON_HEIGHT);
		
		//Re-add drop-down button
		this.add(dropDown);
	}
	
	/**
	 * Returns the number of buttons existing in the combo button.
	 * 
	 * @return the number of buttons
	 */
	public int getButtonCount(){
		return buttons.length;
	}
	
	/**
	 * Returns the array of buttons of type <code>TBButton</code> that
	 * the Combo Button contains
	 * @return array of buttons of type <code>TBButton</code>
	 */
	public TBButton [] getButtons() {
		return buttons;
	}

	/** 
	 * Gets the index of the active button in the list
	 * @return index number inside buttons list (from 0)
	 */
	public int getIndex() {
		return this.index;
	}

	/**
	 * Sets the index of the active button in the list
	 * @param index of the new active button in the list
	 */
	public void setIndex(int index) {
		this.index = index;
	}

	/**
	 * Returns the active tool in the combo list
	 * @return a <code>TBButton</code> with the Active Button
	 */
	public TBButton getActiveTool() {
		return (TBButton) buttons[index];
	}

	/**
	 * Sets the active tool in the combo list
	 * @param _index new index to be the active tool
	 */
	public void setActiveTool(int _index) {
		index = _index;
	}
	
	/**
	 * Sets the active tool in the combo list
	 * @param b new tool to be the active tool
	 */
	public void setActiveTool(TBButton b) {
		
		int sz = buttons.length;
		
		for (int i=0; i<sz; i++) {
			
			if (buttons[i].getHelpText().compareTo(b.getHelpText()) == 0) {
				setActiveTool(i);
				return;
			}
		}

		logger.error("Tool not found in this combo: " + b.getName());
	}

	/**
	 * Override method of <code>java.lang.Object.equals</code> for inspecting equals in the
	 * buttons of the list
	 * @param obj Object to inspect if it's equal
	 * @return boolean <code>true</code> if the object equal any of the buttons
	 * 					in the combo button's button list
	 * @see java.lang.Object.equals
	 */
	public boolean equals(Object obj){
		for (int i=0; i< buttons.length; i++){
			if (buttons[i].equals(obj)) return true;
		}
		return false;
	}
	
	/**
	 * Override method of <code>java.awt.Component.setVisible</code>
	 * It sets visible only the active button and the drop-down button.
	 */
	public void setVisible(boolean b){
		super.setVisible(b);
		if (buttons != null){ 
			buttons[index].setVisible(b);
			dropDown.setVisible(b);
		}
		if (dropDown != null){
			dropDown.setVisible(b);
		}
	}
	
	
	/**
	 * Sets all the buttons in the list to <i>selected/unselected</i> status
	 * @param  _selected <code>true</code>: Set the active button to <i>selected</i> status
	 * 		  <code>false</code>: Sets all the buttons to <i>unselected</i> status
	 */
	public void setSelected(boolean _selected) {
		// logger.debug("TBComboButton: (Un)selecting all buttons");
		if (_selected == false){
			for (int i = 0; i < buttons.length; i++){
				buttons[i].setSelected(_selected);
			}
			dropDown.setSelected(_selected);
		}
		else {
			buttons[index].setSelected(true);
		}
		this.selected = _selected;
	}
	

	/**
	 * @return Returns the dropDown button in the combo button.
	 */
	public TBDropDownButton getDropDown() {
		return dropDown;
	}


	/**
	 * @return Returns the drop-down menu
	 */
	public TBMenu getMenu() {
		return menu;
	}
	
	/**
	 * Returns the MainWindow of the combo button
	 * @return  MainWindow where the toolbar which contains the combo button
	 * 			has been placed
	 */
	public MainWindow getWindow() {
		
		return this.getOwner().getWindow();
	}

	/**
	 * Override method of <code>paint</code> method which forces the paint of
	 * the active button (the visible one in the tool bar) and the arrow 
	 */
	public void paint(Graphics g) {
		buttons[index].paint(g);
		dropDown.paint(g);
	}
	


	public void mouseClicked(MouseEvent e) {
		
		
	}

	public void mouseEntered(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	public void mouseExited(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	public void mousePressed(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	public void mouseReleased(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}



}
