package es.uv.robotica.gvsig.mobile.gps.gui;

import java.awt.Graphics;
import java.awt.Image;
import java.awt.Panel;

/**
 * A double buffer implementation to avoid the flicker.
 *
 * The idea behind double buffering is that you first paint everything
 * to an off-screen image, and, when it's ready, paint it to the screen
 * in just one paintjob.
 *
 * For this purpose, we need a bufferimage and it's brother buffergraphics.
 * This image has always the same dimensions as your panel, so if your
 * panel resizes, the image has to resize also, and the previous has
 * to be cleared out of memory.
 *
 * Subclasses should override the paintBuffer method to paint into the panel.
 *
 * @author zizilamoroso
 * @see http://www.codeproject.com/KB/java/javadoublebuffer.aspx
 */
public class DoubleBufferPanel extends Panel {

	// class variables
	private int bufferWidth;
	private int bufferHeight;
	private Image bufferImage;
	private Graphics bufferGraphics;

	/**
	 * Constructor
	 */
    public DoubleBufferPanel() {
        super();
    }

    /*
     * Override update so that it doesn't clear the panel,
     * which causes the flicker
     */
    public void update(Graphics g) {
        paint(g);
    }

    /*
     * Override paint to continuously update the buffer size with the
     * panel size, and paints the onscreen image in one step.
     * We make it final because we don't want subclasses to override it.
     * Subclasses shall override paintBuffer instead
     */
    public final void paint(Graphics g) {
		// checks the buffersize with the current panel size
		// or initializes the image with the first paint
		if(	bufferWidth!=getSize().width ||
			bufferHeight!=getSize().height ||
			bufferImage==null ||
			bufferGraphics==null)
				resetBuffer();

		if(bufferGraphics!=null){
			// this clears the offscreen image, not the onscreen one
			bufferGraphics.clearRect(0,0,bufferWidth,bufferHeight);

			// calls the paintbuffer method with the offscreen graphics as a param
			paintBuffer(bufferGraphics);

			// we finaly paint the offscreen image onto the onscreen image
			g.drawImage(bufferImage,0,0,this);
		}
    }

	/**
	 * This is the method that subclasses should override to paint the panel.
	 * @param g
	 */
	public void paintBuffer(Graphics g) {
		// in classes extended from this one, add something to paint here!
		// always remember, g is the offscreen graphics
	}

    /*
     * Called the first time the buffer is used and each time the panel
     * is resized to destroy the old buffer and create a new one
     * according to the actual panel size
     */
	private void resetBuffer() {
		// always keep track of the image size
		bufferWidth=getSize().width;
		bufferHeight=getSize().height;

		// clean up the previous image
		if(bufferGraphics!=null){
			bufferGraphics.dispose();
			bufferGraphics=null;
		}
		if(bufferImage!=null){
			bufferImage.flush();
			bufferImage=null;
		}
		// forces garbage collector to work ;)
		System.gc();

		// create the new image with the size of the panel
		bufferImage=createImage(bufferWidth,bufferHeight);
		bufferGraphics=bufferImage.getGraphics();
	}
}
