/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.fmap.driver.raster.wms;

import java.awt.geom.Rectangle2D;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Set;

import org.apache.log4j.Logger;
import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;

/**
 * This class keeps information about a single WMS layer. It is used after
 * a getCapabilities request, to manage capabilities data in a tree-like
 * data model.
 * 
 * @see es.prodevelop.gvsig.mobile.fmap.driver.raster.wms.WMSProtocolHandler
 * @see es.prodevelop.gvsig.mobile.fmap.driver.raster.wms.WmsRasterDriver
 * 
 * @author jldominguez
 *
 */
public class WMSLayerDescription {
	
	private static Logger logger = Logger.getLogger(WMSLayerDescription.class);	
	
	private String lyrId = "";
	private String lyrName = "";
	private HashMap epsgToExtent = new HashMap();
	private ArrayList children = new ArrayList();
	private boolean infoable = false;
	private int level = 0;
	private boolean selected = false;
	private WMSLayerDescription parent = null;
	
	/**
	 * Constructor 
	 * @param lvl depth level of this layer
	 * @param p parent layer (parent node in the tree)
	 */
	public WMSLayerDescription(int lvl, WMSLayerDescription p) {
		parent = p;
		level = lvl;
	}
	
	/**
	 * This method recursively calls to the same method of the node's children.
	 * 
	 * It is used to prevent memory leaks caused by non-destryoed objects, after the
	 * WMS dialog has been closed.
	 *
	 */
	public void delete() {
		for (int i=0; i<children.size(); i++) {
			((WMSLayerDescription) children.get(i)).delete();
		}
		epsgToExtent.clear();
		epsgToExtent = null;
		children.clear();
		children = null;
		parent = null;
	}
	
	/**
	 * Recursively adds useful ID to the provided string. Some layers do not have an ID,
	 * meaning that they are just the parent of other layers.
	 * 
	 * @param str the string to append useful IDS to.
	 * @return a new string that is the result of rescursively
	 * appending useful IDs to the provided string.
	 */
	public String appendUseful(String str) {
		
		String resp = str;
		String id_trim = lyrId.trim();
		int n_child = children.size();
		
		if ((n_child == 0) & (selected) && (id_trim.length() > 0)) {

			resp = resp + "," + id_trim;

		} else {

			WMSLayerDescription item = null;
			for (int i=0; i<n_child; i++) {
				item = (WMSLayerDescription) children.get(i); 
				resp = item.appendUseful(resp);
			}

		}
		return resp;
	}
	
	/**
	 * @return whether the checkbox associated with the layer described hehe
	 * is checked or not. 
	 */
	public boolean isSelected() {
		return selected;
	}
		
	/**
	 * Set the associated checkbox to a value, possibly acting recursively on
	 * children nodes.
	 * 
	 * @param s the new value for the check box.
	 * 
	 * @param tell_children whether the children nodes have to be reset or not.
	 */
	public void setSelected(boolean s, boolean tell_children) {

		selected = s;
		
		if (tell_children) {
			for (int ch = 0; ch<children.size(); ch++) {
				((WMSLayerDescription) children.get(ch)).setSelected(s, true);
			}
		}

		if ((!selected) && (parent != null)) {
			parent.setSelected(false, false);
		}
	}

	/**
	 * 
	 * @return the depth of this layer in the tree that describes the available layers.
	 */
	public int getLevel() {
		return level;
	}
	
	/**
	 * Sets whether or not this layer is infoable (according to the response received
	 * after a get capabilities request.
	 * 
	 * @param b whether or not this layer is infoable
	 */
	public void setInfoable(boolean b) {
		infoable = b;
	}
	
	/**
	 * 
	 * @return whether or not this layer is infoable
	 */
	public boolean isInfoable() {
		return infoable;
	}

	/**
	 * 
	 * @return layer name (not to be confused with layer ID)
	 */
	public String getLayerName() {
		return lyrName;
	}
	
	/**
	 * Sets the layer name
	 * @param n the layer name (not to be confused with layer ID)
	 */
	public void setLayerName(String n) {
		lyrName = n;
	}
	
	/**
	 * 
	 * @return layer ID (not to be confused with layer name)
	 */
	public String getLayerId() {
		return lyrId;
	}
	
	/**
	 * Set the layer ID
	 * @param i the layer ID
	 */
	public void setLayerId(String i) {
		lyrId = i;
	}
	
	/**
	 * 
	 * @return the SRS abbreviations in which this layer is available
	 */
	public String[] getAvailableSRS() {
		
		Set keys = epsgToExtent.keySet();
		Object[] _keys = keys.toArray();
		
		int sz = _keys.length;
		String[] resp = new String[sz];
		for (int i=0; i<sz; i++) {
			resp[i] = (String) _keys[i];  
		}
		return resp;
	}
	
	/**
	 * Gets the extent of this layer for the given SRS
	 * @param srs the abbreviation of the SRS of interest
	 * @return the extent of this layer for the given SRS
	 */
	public Rectangle2D getExtentForSRS(String srs) {
		Object res = epsgToExtent.get(srs);
		
		if (res == null) {
			return null;
		} else {
			return (Rectangle2D) res; 
		}
	}
	
	/**
	 * Adds a child to this layer description
	 * (used while the capabilities response is being parsed)
	 * @param c the new child to be added
	 */
	public void addChild(WMSLayerDescription c) {
		children.add(c);
	}

	/**
	 * Adds a new SRS, indicating that this layer is also available in that SRS
	 * @param srs the new SRS to be added
	 * @param ext the extent of the layer in the provided SRS
	 */
	public void addCRS(String srs, Rectangle2D ext) {
		epsgToExtent.put(srs, ext);
	}
	
	/**
	 * Increase the extent of this layer to include the provided rectangle.
	 * 
	 * @param srs the SRS that the extent refers to
	 * @param ext the new rectangle to be included
	 * in the extent associated with the provided SRS
	 */
	public void increaseExtentForSRS(String srs, Rectangle2D ext) {
		
		Rectangle2D curr_ext = (Rectangle2D) epsgToExtent.get(srs);
		
		if (curr_ext == null) {
			addCRS(srs, ext);
		} else {
			Rectangle2D.Double union = new Rectangle2D.Double(0, 0, 1, 1);
			Rectangle2D.union(curr_ext, ext, union);
			addCRS(srs, union);
		}
	}

	/**
	 * Clears the list of available SRS
	 *
	 */
	public void clearSRS() {
		epsgToExtent.clear();
	}

	/**
	 * This method recursively parses the layers section of the capabilities response.
	 * Data is stored in the same object that performs the parsing.
	 * 
	 *  
	 * @param parser the parser 
	 * @param layer_found whether a layer description is right
	 * ahead of the current position of the parser
	 * @throws IOException
	 * @throws XmlPullParserException
	 */
	public void parse(KXmlParser parser, boolean layer_found) throws IOException, XmlPullParserException {

        int currentTag;
        boolean end = false;
        String value;
        
        if (!layer_found)
        	parser.require(KXmlParser.START_TAG, null, CapabilitiesTags.LAYER);
       
        readLayerAttributes(parser);
        currentTag = parser.nextTag();
        
        while (!end) 
        {
            switch(currentTag)
            {
                case KXmlParser.START_TAG:
                    if (parser.getName().compareTo(CapabilitiesTags.LAYER)==0)
                    {	
                    	WMSLayerDescription lyr = new WMSLayerDescription(level + 1, this);						
                        lyr.parse(parser, true);
                        addChild(lyr);
                    }
                    else if (parser.getName().compareTo(CapabilitiesTags.ATTRIBUTION)==0){
                        // TODO comprobar que esto se necesite o se deseche
                        parser.skipSubTree();
                    }
                    else if (parser.getName().compareTo(CapabilitiesTags.NAME)==0)
                    {		
                        value = parser.nextText();
                        if (value != null) {
                        	lyrId = value;						
                        }
                    }	
                    else if (parser.getName().compareTo(CapabilitiesTags.TITLE)==0)
                    {
                        value = parser.nextText();
                        if (value != null) lyrName = value;
                    }
                    
                    /*
                    else if (parser.getName().compareTo(CapabilitiesTags.ABSTRACT)==0)
                    {
                        value = parser.nextText();
                        if (value != null) setAbstract(value);
                    }
                    
                    else if (parser.getName().compareTo(CapabilitiesTags.SRS)==0)
                    {
                        value = parser.nextText();
                        if (value != null){
                            String[] mySRSs = value.split(" ");
                            for (int i = 0; i < mySRSs.length; i++) {
                                addSrs(mySRSs[i]);    
                            }                            
                        }
                    }
                    */
                    
                    else if (parser.getName().compareTo(CapabilitiesTags.BOUNDINGBOX)==0)
                    {
                    	double minx = 0;
                    	double maxx = 0;
                    	double miny = 0;
                    	double maxy = 0;
                    	String _srs = "";
                    	
                    	try {
                            value = parser.getAttributeValue("",CapabilitiesTags.SRS);
                            _srs = value;
                        	
                        	if (_srs != null) {
                                value = parser.getAttributeValue("",CapabilitiesTags.MINX);
                                if (value != null) minx = Double.parseDouble(value);
                                value = parser.getAttributeValue("",CapabilitiesTags.MAXX);
                                if (value != null) maxx = Double.parseDouble(value);
                                value = parser.getAttributeValue("",CapabilitiesTags.MINY);
                                if (value != null) miny = Double.parseDouble(value);
                                value = parser.getAttributeValue("",CapabilitiesTags.MAXY);
                                if (value != null) maxy = Double.parseDouble(value);

                            	Rectangle2D ext = new Rectangle2D.Double(
                            			minx, miny, (maxx - minx), (maxy - miny));
                            	epsgToExtent.put(_srs, ext);
                        	}
                    	} catch (Exception ex) {
                    		logger.error("While reading extent, srs = " + _srs + ", msg = " + ex.getMessage());
                    	}
                    }	
                    else if (parser.getName().compareTo(CapabilitiesTags.LATLONBOUNDINGBOX)==0)
                    {
                    	double minx = 0;
                    	double maxx = 0;
                    	double miny = 0;
                    	double maxy = 0;

                    	try {
                                value = parser.getAttributeValue("",CapabilitiesTags.MINX);
                                if (value != null) minx = Double.parseDouble(value);
                                value = parser.getAttributeValue("",CapabilitiesTags.MAXX);
                                if (value != null) maxx = Double.parseDouble(value);
                                value = parser.getAttributeValue("",CapabilitiesTags.MINY);
                                if (value != null) miny = Double.parseDouble(value);
                                value = parser.getAttributeValue("",CapabilitiesTags.MAXY);
                                if (value != null) maxy = Double.parseDouble(value);

                            	Rectangle2D ext = new Rectangle2D.Double(
                            			minx, miny, (maxx - minx), (maxy - miny));
                            	epsgToExtent.put("EPSG:4326", ext);
                    	} catch (Exception ex) {
                    		logger.error("While reading lat lon extent: " + ex.getMessage());
                    	}
                    }						
                    /*
                    else if (parser.getName().compareTo(CapabilitiesTags.SCALEHINT)==0)
                    {
                        value = parser.getAttributeValue("",CapabilitiesTags.MIN);
                        if ((value != null) && (Utilities.isNumber(value)))
                            setScaleMin(Double.parseDouble(value));
                        value = parser.getAttributeValue("",CapabilitiesTags.MAX);
                        if ((value != null) && (Utilities.isNumber(value)))
                            setScaleMax(Double.parseDouble(value));																	
                    }						
                    */
                    else if (parser.getName().compareTo(CapabilitiesTags.STYLE)==0)
                    {
                        // TODO comprobar que esto se necesite o se deseche
                        parser.skipSubTree();
                    }
                    /*
                    else if (parser.getName().compareTo(CapabilitiesTags.DIMENSION)==0)
                    {
                        WMSDimension dim = new WMSDimension();
                        dim.parse(parser);
                        if ((dim != null) && (dim.getName() != null))
                        {
                            addDimension(dim);
                        }
                    }
                    else if (parser.getName().compareTo(CapabilitiesTags.EXTENT)==0)
                    {                    	
                        WMSExtent extent = new WMSExtent();
                        extent.parse(parser);
                        if ((extent != null) && (extent.getName() != null))
                        {
                            addExtent(extent);
                            
                        }
                    }                      
                    else if (parser.getName().compareTo(CapabilitiesTags.KEYWORDLIST)==0)
                    {
                    	parseKeywordList(parser);
                    } 
                    */                   
                    break;
                case KXmlParser.END_TAG:
                    if (parser.getName().compareTo(CapabilitiesTags.LAYER) == 0)
                        end = true;
                    break;
                case KXmlParser.TEXT:					
                    break;
            }
            if (!end)
            	currentTag = parser.next();
        }
        parser.require(KXmlParser.END_TAG, null, CapabilitiesTags.LAYER);
	}
	
	/**
	 * Starts the parsing process assuming that this object represents
	 * the root layer. Its name will be "[WMS]" and it will have no ID. It's
	 * simply a convenience node because sometimes the server does not
	 * explicitly define a root node.
	 * 
	 * @param parser the parser
	 * @return the code of the tag after the parsing ends. Tghis is useful to
	 * perform coherent parsings.
	 * 
	 * @throws IOException
	 * @throws XmlPullParserException
	 */
	public int parseAsRoot(KXmlParser parser) throws IOException, XmlPullParserException {

		lyrId = "";		
		lyrName = "[WMS]";
		
        int currentTag;
        boolean end = false;

        currentTag = KXmlParser.START_TAG;
        
    	// several layers 
        while ((currentTag == KXmlParser.START_TAG)
        		&& (parser.getName().compareTo(CapabilitiesTags.LAYER)==0)) {
        	
        	end = false;
            while (!end) 
            {
                switch(currentTag)
                {
                    case KXmlParser.START_TAG:
                        if (parser.getName().compareTo(CapabilitiesTags.LAYER)==0)
                        {	
                        	WMSLayerDescription lyr = new WMSLayerDescription(level + 1, this);						
                            lyr.parse(parser, true);
                            addChild(lyr);
                        }
                    case KXmlParser.END_TAG:
                        if (parser.getName().compareTo(CapabilitiesTags.LAYER) == 0)
                            end = true;
                        break;
                }
                currentTag = parser.next();
            }
        	

            // next element, possibly another layer
            while (isIrrelevantTag(currentTag)) currentTag = parser.next();
        }
        
        logger.debug("No more layers -");
        return currentTag;
	}
	
	
    private boolean isIrrelevantTag(int t) {
    	return t == KXmlParser.TEXT;
	}

	/**
     * Reads and parses the layer attributes
     * Maybe this method should be moved to the WMSLayer.
     * Until now the attributes are the same for all versions.
     * @param parser
     */
    protected void readLayerAttributes(KXmlParser parser)
    {
    	String value = new String();

        //First of all set whether the layer is Queryable reading the attribute.
        value = parser.getAttributeValue("", CapabilitiesTags.QUERYABLE);
        if (value != null)
        {
            if (value.compareTo("0") != 0) infoable = true;
        }
        /*
        value = parser.getAttributeValue("", CapabilitiesTags.OPAQUE);
        if (value != null)
        {
            if (value.compareTo("0")==0)
                setOpaque(false);
            else
                setOpaque(true);
        }
        value = parser.getAttributeValue("", CapabilitiesTags.NOSUBSETS);
        if (value != null)
        {
            if (value.compareTo("0")==0)
                setNoSubSets(false);
            else
            	setNoSubSets(true);
        }
        value = parser.getAttributeValue("", CapabilitiesTags.FIXEDWIDTH);
        if (value != null)
        {
        	setfixedWidth(Integer.parseInt(value));
        }
        value = parser.getAttributeValue("", CapabilitiesTags.FIXEDHEIGHT);
        if (value != null)
        {
        	setfixedHeight(Integer.parseInt(value));
        }
        */
    }
    
    /**
     * 
     * @return number of children of this layer description in the tree-like structure
     */
    public int getChildrenCount() {
    	return children.size();
    }
    
    /**
     * Gets child description by index
     * @param i the idnex of the child of interest
     * @return child description by index
     */
    public WMSLayerDescription getChild(int i) {
    	if (i >= getChildrenCount()) {
    		return null;
    	}
    	return (WMSLayerDescription) children.get(i);
    	
    }
    
}
