/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.fmap.driver.raster.wms;

import java.awt.geom.Rectangle2D;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.URL;
import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;

import es.prodevelop.gvsig.mobile.fmap.util.download.DownloadTask;
import es.prodevelop.gvsig.mobile.fmap.util.download.Downloader;
import es.prodevelop.gvsig.mobile.fmap.util.string.StringUtilities;

/**
 * 
 * This is the class that interacts with the WMS drivers. Sends requests
 * (capabilities, images, etc) and
 * starts the parsing process. It also instantiates the root of the layer descriptions
 * tree, and provides a getter used by the WMS dialog to show the results.
 * 
 * @see es.prodevelop.gvsig.mobile.fmap.driver.raster.wms.WmsRasterDriver
 * @see es.prodevelop.gvsig.mobile.fmap.driver.raster.wms.WMSLayerDescription
 * 
 * @author iver
 * @author jldominguez
 *
 */
public class WMSProtocolHandler {

	private static Logger logger = Logger.getLogger(WMSProtocolHandler.class);

	public static final String DEFAULT_WMS_VERSION = "1.1.1";
	private String encoding = "UTF-8";
	private String host;
	// private String port;
	private String version = "";
	private String symbol = "";

	public WMSLayerDescription root;
	private boolean infoable = false;
	private ArrayList formats = new ArrayList();
	private ArrayList infoFormats = new ArrayList();

	/**
	 * @return the root of the layer descriptions tree
	 */
	public WMSLayerDescription getRoot() {
		return root;
	}

	/**
	 * Removes all the objects owned by this object. This is used when the WMS dialog closes,
	 * in order to prevent memory leaks. 
	 *
	 */
	public void delete() {

		if (root != null) root.delete();
		formats.clear();
		formats = null;
		infoFormats.clear();
		infoFormats = null;
	}
	
	/**
	 * Sets the WMS protocol version (1.1.1, etc)
	 * 
	 * @param v the WMS protocol version
	 */
	public void setVersion(String v) {
		version = v;
	}
	
	/**
	 * Sets the path separator symbol to use with the current WMS server
	 * @param s the path separator symbol to use with the current WMS server
	 */
	public void setSymbol(String s) {
		symbol = s;
	}
	
	/**
	 * 
	 * @return the WMS protocol version (1.1.1, etc)
	 */
	public String getVersion() {
		return version;
	}
	
	/**
	 * 
	 * @return the path separator symbol to use with the current WMS server
	 */
	public String getSymbol() {
		return symbol;
	}
	

	/**
	 * 
	 * @return an array with the image formats supoprted by the current WMS server
	 */
	public ArrayList getFormats() {
		return formats;
	}

	/**
	 * 
	 * @return an array with the image formats supported by the current WMS server
	 */
	public ArrayList getInfoFormats() {
		return infoFormats;
	}

	/**
	 * 
	 * @return whether the layer described by this object is declared as infoable by
	 * the server in the getcapabilities response
	 */
	public boolean isInfoable() {
		return infoable;
	}

	    
	/**
	 * 
	 * @return the host url
	 */
	public String getHost() {
		return host;
	}

	/**
	 * Sets the host url
	 * @param _host the new host url
	 */
	public void setHost(String _host) {
		host = _host;
	}

	/**
	 * This method starts the get capabilities request process and parses the result.
	 * 
	 * @param version the WMS protocol version to be used
	 * @param sym the path separator symbol to be used when building requests
	 * @return whether the get capabilities request was completed successfully
	 */
	public boolean getCapabilities(String version, String sym) {

		URL request = null;
		try {
			request = new URL(buildCapabilitiesRequest(host, version, sym));
		} catch (Exception e) {
			logger.error("Bad capabilities request URL: " + e.getMessage());
		}

		try {
			// File f = Downloader.get(request, "cap");

			File f = (File) Downloader.downloadObjectCancel(request, "cap",
					DownloadTask.FILE, "wms", null, true);

			if (f == null) {
				logger.error("capabilities file is null");
				return false;
			}
			clear();
			parse(f);
		} catch (Exception e) {
			logger.error("While downloading/parsing WMS capabilities file: "
					+ e.getMessage());
			return false;
		}

		// in case root has no SRS:
		avoidEmptySRSInRoot();

		return true;
	}

	private void avoidEmptySRSInRoot() {

		if (root == null)
			return;
		if (root.getAvailableSRS().length > 0)
			return;

		logger
				.debug("Root layer has no SRS array. Those of its first-level children will be added...");

		int cnt = root.getChildrenCount();
		for (int i = (cnt - 1); i >= 0; i--) {
			WMSLayerDescription desc = root.getChild(i);
			String[] ava_srs = desc.getAvailableSRS();
			for (int j = 0; j < ava_srs.length; j++) {
				String item_srs = ava_srs[j];
				Rectangle2D ext = desc.getExtentForSRS(item_srs);
				root.increaseExtentForSRS(item_srs, ext);
			}
		}
	}

	private String buildCapabilitiesRequest(String h, String v, String sym) {
		return h + sym + "REQUEST=GetCapabilities&SERVICE=WMS&VERSION=" + v
				+ "&EXCEPTIONS=XML";
	}

	private void clear() {
		infoable = false;
		root = new WMSLayerDescription(0, null);
		formats.clear();
		infoFormats.clear();
	}

	protected String parseException(byte[] data) {
		ArrayList errors = new ArrayList();
		KXmlParser kxmlParser = new KXmlParser();
		try {
			kxmlParser.setInput(new ByteArrayInputStream(data), encoding);
			kxmlParser.nextTag();
			int tag;
			if (kxmlParser.getEventType() != KXmlParser.END_DOCUMENT) {
				kxmlParser.require(KXmlParser.START_TAG, null,
						EXCEPTION_TAGS_EXCEPTION_ROOT);
				tag = kxmlParser.nextTag();
				while (tag != KXmlParser.END_DOCUMENT) {
					switch (tag) {
					case KXmlParser.START_TAG:
						if (kxmlParser.getName().compareTo(
								EXCEPTION_TAGS_SERVICE_EXCEPTION) == 0) {
							String errorCode = kxmlParser.getAttributeValue("",
									EXCEPTION_TAGS_CODE);
							errorCode = (errorCode != null) ? "[" + errorCode
									+ "] " : "";
							String errorMessage = kxmlParser.nextText();
							errors.add(errorCode + errorMessage);
						}
						break;
					case KXmlParser.END_TAG:
						break;

					}
					tag = kxmlParser.nextTag();
				}
				//kxmlParser.require(KXmlParser.END_DOCUMENT, null, null);
			}
		} catch (XmlPullParserException parser_ex) {
			parser_ex.printStackTrace();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
		String message = errors.size() > 0 ? "" : null;
		for (int i = 0; i < errors.size(); i++) {
			message += (String) errors.get(i) + "\n";
		}
		return message;
	}

	/**
	 * Closes the protocol handler. Currently empty.
	 *
	 */
	public void close() {
		// your code here
	}

	/**
	 * Tag used when parsing WMS responses 
	 */
	public final static String EXCEPTION_TAGS_EXCEPTION_ROOT = "ServiceExceptionReport";
	/**
	 * Tag used when parsing WMS responses 
	 */
	public final static String EXCEPTION_TAGS_SERVICE_EXCEPTION = "ServiceException";
	/**
	 * Tag used when parsing WMS responses 
	 */
	public final static String EXCEPTION_TAGS_CODE = "code";

	/**
	 * Main parse method.
	 * 
	 * @param f file to be parsed
	 */
	private void parse(File f) {

		FileReader reader = null;
		try {
			reader = new FileReader(f);

			//	    		FileInputStream fis = new FileInputStream(f);
			//	    		InputStreamReader reader = new InputStreamReader(fis,"ISO-8859-1");
			// BufferedReader br = new BufferedReader(reader);

			char[] buffer = new char[100];
			reader.read(buffer);
			StringBuffer strBuff = new StringBuffer();
			strBuff.append(buffer);
			String auxStr = strBuff.toString().toLowerCase();
			int a = auxStr.toLowerCase().indexOf("<?xml");
			if (a != -1)
				auxStr = auxStr.substring(a, auxStr.length());

			StringBuffer st = new StringBuffer(auxStr);
			String searchText = "encoding=\"";
			int index = st.indexOf(searchText);

			if (index > -1) {
				st.delete(0, index + searchText.length());
				encoding = st.substring(0, st.indexOf("\""));
			}
			
			reader.close();
			
		} catch (Exception ex) {
			logger.error("In main parse method: " + ex.getMessage());
		}

		int tag;
		KXmlParser kxmlParser = null;
		kxmlParser = new KXmlParser();
		try {
//			reader = new FileReader(f);
//			//	    		FileInputStream fis = new FileInputStream(f);
//			//	    		InputStreamReader reader = new InputStreamReader(fis,"ISO-8859-1");
//			// BufferedReader br = new BufferedReader(reader);
//			// patch for ArcIMS + WMS connector > 9.0 bug
//			
//			char[] buffer = new char[(int) f.length()];
//			
//			logger.debug("PHONEME - buffer.len = " + buffer.length);
//			
//			reader.read(buffer);
//			
//			StringBuffer strBuff = new StringBuffer();
//			strBuff.append(buffer);
//			String auxStr = strBuff.toString();
//			
//			logger.debug("PHONEME - auxStr = " + auxStr);
//			// String string = new String(buffer);
//			
//			int a = auxStr.toLowerCase().indexOf("<?xml");
//			
//			logger.debug("PHONEME - a = " + a);
//			
//			if (a != -1) {
//				auxStr = auxStr.substring(a, auxStr.length());
//				logger.debug("PHONEME - auxStr = " + auxStr);
//				kxmlParser.setInput(new StringReader(auxStr));
//			} else
//				// end patch
//				kxmlParser.setInput(new FileInputStream(f), encoding);

			
			FileInputStream fir = new FileInputStream(f);
			kxmlParser.setInput(fir, encoding);
			
			
			logger.debug("PHONEME - kxmlParser.nextTag();");
			
			kxmlParser.nextTag();
			if (kxmlParser.getEventType() != KXmlParser.END_DOCUMENT) {
				kxmlParser.require(KXmlParser.START_TAG, null,
						CapabilitiesTags.CAPABILITIES_ROOT1_1_1);
				tag = kxmlParser.nextTag();
				while (tag != KXmlParser.END_DOCUMENT) {
					switch (tag) {

					case KXmlParser.START_TAG:
						if (kxmlParser.getName().compareTo(
								CapabilitiesTags.SERVICE) == 0) {
							// parseServiceTag(kxmlParser);
						} else if (kxmlParser.getName().compareTo(
								CapabilitiesTags.CAPABILITY) == 0) {
							parseCapabilityTag(kxmlParser);
						}
						break;
					case KXmlParser.END_TAG:
						break;
					case KXmlParser.TEXT:
						break;
					}
					tag = kxmlParser.next();
				}
				kxmlParser.require(KXmlParser.END_DOCUMENT, null, null);
				fir.close();
			}
		} catch (XmlPullParserException parser_ex) {
			logger.error("Parser exception: " + parser_ex.getMessage());
		} catch (IOException ioe) {
			logger.error("IO exception: " + ioe.getMessage());
		} finally {

		}
		// In the parsing process the layer has been filled
	}

	/**
	 * <p>Parses the Service Information </p>
	 */

	/*
	 private void parseServiceTag(KXmlParser parser) throws IOException, XmlPullParserException
	 {
	 int currentTag;
	 boolean end = false;

	 parser.require(KXmlParser.START_TAG, null, CapabilitiesTags.SERVICE);
	 currentTag = parser.next();

	 while (!end)
	 {
	 switch(currentTag)
	 {
	 case KXmlParser.START_TAG:
	 if (parser.getName().compareTo(CapabilitiesTags.NAME)==0)
	 {
	 serviceInfo.name = parser.nextText();
	 }
	 else if (parser.getName().compareTo(CapabilitiesTags.TITLE)==0)
	 {
	 serviceInfo.title = parser.nextText();
	 }
	 else if (parser.getName().compareTo(CapabilitiesTags.ABSTRACT)==0)
	 {
	 serviceInfo.abstr = parser.nextText();
	 }
	 else if (parser.getName().compareTo(CapabilitiesTags.ONLINERESOURCE)==0)
	 {
	 String value = new String();
	 value = parser.getAttributeValue("", CapabilitiesTags.XLINK_HREF);
	 if (value != null){
	 serviceInfo.online_resource = value;
	 }
	 }
	 else if ((parser.getName().compareTo(CapabilitiesTags.KEYWORDLIST)==0) ||
	 (parser.getName().compareTo(CapabilitiesTags.CONTACTINFORMATION)==0))
	 {
	 parser.skipSubTree();
	 }
	 break;
	 case KXmlParser.END_TAG:
	 if (parser.getName().compareTo(CapabilitiesTags.SERVICE) == 0)
	 end = true;
	 break;
	 case KXmlParser.TEXT:
	 break;
	 }
	 if (!end)
	 currentTag = parser.next();
	 }
	 parser.require(KXmlParser.END_TAG, null, CapabilitiesTags.SERVICE);
	 }
	 */

	/**
	 * <p>Parses the Capability Tag </p>
	 */
	private void parseCapabilityTag(KXmlParser parser) throws IOException,
			XmlPullParserException {
		int currentTag;
		boolean end = false;

		parser.require(KXmlParser.START_TAG, null, CapabilitiesTags.CAPABILITY);
		currentTag = parser.next();

		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().compareTo(CapabilitiesTags.REQUEST) == 0) {
					parseRequestTag(parser);
				} else if (parser.getName().compareTo(
						CapabilitiesTags.EXCEPTION) == 0) {
					//TODO:
					//Add to serviceInfo the supported formats for the exceptions????
				} else if (parser.getName().compareTo(CapabilitiesTags.LAYER) == 0) {
					int curr_tag = root.parseAsRoot(parser);
					end = ((parser.getName().compareTo(
							CapabilitiesTags.CAPABILITY) == 0) && (curr_tag == KXmlParser.END_TAG));
				} else if ((parser.getName().compareTo(
						CapabilitiesTags.VENDORSPECIFICCAPABILITIES) == 0)
						|| (parser.getName().compareTo(
								CapabilitiesTags.USERDEFINEDSYMBOLIZATION) == 0)
						|| (parser.getName().compareTo(CapabilitiesTags.STYLE) == 0))

				{
					parser.skipSubTree();
				}
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(CapabilitiesTags.CAPABILITY) == 0)
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
		//parser.require(KXmlParser.END_TAG, null, CapabilitiesTags.CAPABILITY);
	}

	/**
	 * <p>Parses the Request tag </p>
	 */
	private void parseRequestTag(KXmlParser parser) throws IOException,
			XmlPullParserException {
		int currentTag;
		boolean end = false;

		parser.require(KXmlParser.START_TAG, null, CapabilitiesTags.REQUEST);
		currentTag = parser.next();

		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().compareTo(CapabilitiesTags.GETMAP) == 0) {
					parseGetMapTag(parser);
				} else if (parser.getName().compareTo(
						CapabilitiesTags.GETFEATUREINFO) == 0) {
					infoable = true;
					parseGetFeatureInfoTag(parser);
				}
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(CapabilitiesTags.REQUEST) == 0)
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
		// TODO: does not get such a tag when arrives here!!!!!!
		//parser.require(KXmlParser.END_TAG, null, CapabilitiesTags.REQUEST);
	}

	/**
	 * <p>Parses the GetMap tag </p>
	 */
	private void parseGetMapTag(KXmlParser parser) throws IOException,
			XmlPullParserException {
		int currentTag;
		boolean end = false;

		parser.require(KXmlParser.START_TAG, null, CapabilitiesTags.GETMAP);
		currentTag = parser.next();

		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().compareTo(CapabilitiesTags.FORMAT) == 0) {
					String f = parser.nextText();
					if (validFormat(f)) {
						formats.add(f);
					}

				}
				/*
				 else if (parser.getName().compareTo(CapabilitiesTags.DCPTYPE)==0)
				 {
				 currentTag = parser.nextTag();
				 if(parser.getName().compareTo(CapabilitiesTags.HTTP)==0)
				 {
				 currentTag = parser.nextTag();
				 if(parser.getName().compareTo(CapabilitiesTags.GET)==0)
				 {
				 currentTag = parser.nextTag();
				 if (parser.getName().compareTo(CapabilitiesTags.ONLINERESOURCE)==0)
				 {
				 String value = new String();
				 value = parser.getAttributeValue("", CapabilitiesTags.XLINK_HREF);
				 if (value != null){
				 serviceInfo.operations.put(CapabilitiesTags.GETMAP, value);
				 }
				 }
				 }
				 }
				 }
				 */
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(CapabilitiesTags.GETMAP) == 0)
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
	}

	private boolean validFormat(String f) {

		String format = f.toLowerCase();
		if (format.indexOf("jpg") != -1)
			return true;
		if (format.indexOf("jpeg") != -1)
			return true;
		if (format.indexOf("gif") != -1)
			return true;
		if (format.indexOf("png") != -1)
			return true;
		return false;
	}

	/**
	 * <p>Parses the GetFeatureInfoTag tag </p>
	 */

	private void parseGetFeatureInfoTag(KXmlParser parser) throws IOException,
			XmlPullParserException {
		int currentTag;
		boolean end = false;

		parser.require(KXmlParser.START_TAG, null,
				CapabilitiesTags.GETFEATUREINFO);
		currentTag = parser.next();

		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().compareTo(CapabilitiesTags.FORMAT) == 0) {
					//TODO:
					// add the supported formats by the GetFeatureInfo request
					infoFormats.add(parser.nextText());
					//serviceInfo.formats.add(parser.nextText());
				}
				/*
				 else if (parser.getName().compareTo(CapabilitiesTags.DCPTYPE)==0)
				 {
				 currentTag = parser.nextTag();
				 if(parser.getName().compareTo(CapabilitiesTags.HTTP)==0)
				 {
				 currentTag = parser.nextTag();
				 if(parser.getName().compareTo(CapabilitiesTags.GET)==0)
				 {
				 currentTag = parser.nextTag();
				 if (parser.getName().compareTo(CapabilitiesTags.ONLINERESOURCE)==0)
				 {
				 String value = new String();
				 value = parser.getAttributeValue("", CapabilitiesTags.XLINK_HREF);
				 if (value != null){
				 serviceInfo.operations.put(CapabilitiesTags.GETFEATUREINFO, value);
				 }
				 }
				 }
				 }
				 }
				 */
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(CapabilitiesTags.GETFEATUREINFO) == 0)
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
	}

	/**
	 * Utility method to find out the convenient path separator symbol for a given url string.
	 * @param url the URL of interest
	 * @return the convenient path separator symbol
	 */
	public static String getSymbol(String url) {

		String symbol;

		if (url.indexOf("?") == -1)
			symbol = "?";
		else if (url.indexOf("?") != url.length() - 1)
			symbol = "&";
		else
			symbol = "";

		return symbol;
	}

	private static String getVersionQueryRequest(String _host) {

		String req = "";
		String symbol = getSymbol(_host);
		req = req + _host + symbol + "REQUEST=GetCapabilities&SERVICE=WMS&";
		req += ("&EXCEPTIONS=XML");
		return req;
	}

	/**
	 * Sends a GetCapabilities to the WMS server to get the version
	 * if the version parameter is null, the WMS will return the highest version supported
	 * if not it will return the lower highest version than the one requested.
	 * @param host host URL
	 * @return suitable version supported by the server 
	 */
	public static String getSuitableWMSVersion(URL host)
			throws BadWMSResponseException {

		String request = getVersionQueryRequest(host.toString());
		String version = "";

		byte[] version_bytes = null;

		try {

			URL req_url = new URL(request);

			version_bytes = (byte[]) Downloader.downloadObjectCancel(req_url,
					"", DownloadTask.BYTE_ARRAY, "", "", false);

		} catch (Exception ex) {

			logger.error("While getting version request response: "
					+ ex.getMessage());
			throw new BadWMSResponseException(ex.getMessage());
		}

		String response = new String(version_bytes);

		int a = response.toLowerCase().indexOf("<?xml");
		if (a != -1) {
			response = response.substring(a, response.length());
		}

		StringReader reader = new StringReader(response);
		KXmlParser kxmlParser = null;
		kxmlParser = new KXmlParser();

		try {
			kxmlParser.setInput(reader);
			kxmlParser.nextTag();

			if (kxmlParser.getEventType() != KXmlParser.END_DOCUMENT) {
				if ((kxmlParser.getName().compareTo(
						CapabilitiesTags.CAPABILITIES_ROOT1_1_0) == 0)
						|| (kxmlParser.getName().compareTo(
								CapabilitiesTags.CAPABILITIES_ROOT1_1_1) == 0)
						|| (kxmlParser.getName().compareTo(
								CapabilitiesTags.CAPABILITIES_ROOT1_3_0) == 0)) {
					version = kxmlParser.getAttributeValue("",
							CapabilitiesTags.VERSION);
				}
			}
			// do not forget to close the Stream.    		
			reader.close();

			if (version.length() == 0) {
				throw new BadWMSResponseException("No WMS version found.");
			}

			return version;

		} catch (Exception ex) {
			logger.error("While parsing version request response: "
					+ ex.getMessage());
			throw new BadWMSResponseException(ex.getMessage());
		} finally {
			if (reader != null) {
				reader.close();
			}
		}
	}

	/**
	 * This method excludes from a comma separated string of IDs, the ones that are not infoable.
	 * It is used for querying only on queriable layers, and prevents empty responses on some
	 * WMS servers.
	 *  
	 * @param all_csv comma separated string of all layers requested by the user
	 * @param root the layer description object of the root layer
	 * @param getname whether we wanrt to get the names or the IDs of the layers
	 * @return a comma separated string with the IDs or names of the queryable layers
	 */
	public static String leaveInfoable(String all_csv,
			WMSLayerDescription root, boolean getname) {

		String[] all_array = StringUtilities.split(all_csv, ",");
		ArrayList resp = new ArrayList();
		addRecursiveYourselfIfPresent(root, all_array, resp, getname);

		String _resp = null;
		int sz = resp.size();
		if (sz > 0) {
			_resp = "";
			for (int i = 0; i < sz; i++)
				_resp = _resp + "," + ((String) resp.get(i));
			_resp = _resp.substring(1);
		}
		return _resp;

	}

	private static void addRecursiveYourselfIfPresent(WMSLayerDescription desc,
			String[] sel, ArrayList resp, boolean getname) {

		if (desc.isInfoable() && isOneOf(desc.getLayerId(), sel)) {
			if (getname) {
				// name
				resp.add(desc.getLayerName());
			} else {
				// id
				resp.add(desc.getLayerId());
			}

		}

		for (int i = 0; i < desc.getChildrenCount(); i++) {
			addRecursiveYourselfIfPresent(desc.getChild(i), sel, resp, getname);
		}
	}
	
	
	
	private static void addRecursiveYourselfIfPresentNew(WMSLayerDescription desc,
			String[] sel, ArrayList resp, boolean getname) {

		if (isOneOf(desc.getLayerId(), sel)) {
			
			if (desc.isInfoable()) {
				if (getname) {
					resp.add(desc.getLayerName()); // name
				} else {
					resp.add(desc.getLayerId()); // id
				}
			}
			addAllInfoableChildren(desc, resp, getname);

		} else {
			
			// ---------- search children
			for (int i = 0; i < desc.getChildrenCount(); i++) {
				addRecursiveYourselfIfPresentNew(desc.getChild(i), sel, resp, getname);
			}
			
		}

	}
	
	private static void addAllInfoableChildren(
			WMSLayerDescription desc,
			ArrayList resp,
			boolean getname) {

		for (int i = 0; i < desc.getChildrenCount(); i++) {

			WMSLayerDescription ch = desc.getChild(i);
			if (ch.isInfoable()) {
				if (getname) {
					resp.add(ch.getLayerName());
				} else {
					resp.add(ch.getLayerId());
				}
			}
			addAllInfoableChildren(ch, resp, getname);
		}
	}

	private static boolean isOneOf(String str, String[] arr) {

		for (int i = 0; i < arr.length; i++) {
			if (str.compareTo(arr[i]) == 0)
				return true;
		}
		return false;

	}

	public static String firstLayerName(String[] sel, WMSLayerDescription root) {
		return getFirstPresent(root, sel);
	}
	
	private static String getFirstPresent(WMSLayerDescription desc, String[] sel) {

		if (isOneOf(desc.getLayerId(), sel)) {
			return desc.getLayerName();
		}

		for (int i = 0; i < desc.getChildrenCount(); i++) {
			String resp = getFirstPresent(desc.getChild(i), sel);
			if (resp != null) {
				return resp; 
			}
		}
		
		return null;
	}

}

