/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 Prodevelop and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *   +34 963862235
 *   gvsig@gva.es
 *   http://www.gvsig.gva.es
 *
 *    or
 *
 *   Prodevelop Integracin de Tecnologas SL
 *   Conde Salvatierra de lava , 34-10
 *   46004 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *
 *    or
 *
 *   Instituto de Robtica
 *   Apartado de correos 2085
 *   46071 Valencia
 *   (Spain)
 *   
 *   +34 963 543 577
 *   jjordan@robotica.uv.es
 *   http://robotica.uv.es
 *   
 */

package es.prodevelop.gvsig.mobile.fmap.util.download;

import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.net.HttpURLConnection;
import java.net.URL;
import java.nio.ByteBuffer;
import java.util.ArrayList;

import org.apache.log4j.Logger;

import es.prodevelop.gvsig.mobile.fmap.util.ResourceReader;
import es.prodevelop.gvsig.mobile.fmap.util.Utils;


/**
 * Utility class to download resources from the internet.
 * 
 * @see es.prodevelop.gvsig.mobile.fmap.util.download.DownloadTask
 * @see es.prodevelop.gvsig.mobile.fmap.util.download.DownloadException
 * 
 * @author jldominguez
 *
 */
public class Downloader {
	
	private static final long MAX_TIME_ALLOWED = 12000;
	private static final long MAX_TIME_ALLOWED_STEP = 500;
	
	public static String CACHED_FILES_ROOT_PATH = File.separator;
	
	private static Logger logger = Logger.getLogger(Downloader.class);
	

	/**
	 * This method downloads the resource into a local File or as an array of bytes in memory.
	 * Returns local file if exists.
	 * 
	 * @param u the URL
	 * @param exte the extension (used if it's going to be a file)
	 * @param type download type (file/array)
	 * @param cachedFilesSubfolder folder destination (used when it's a file)
	 * @param filePath file path on the folder destination
	 * @param use_hash whether a hash function must be used to code the file name
	 * @return an object file or an array of bytes
	 * @throws DownloadException
	 */
	public static Object downloadObjectCancel(
			URL u,
			String exte,
			int type,
			String cachedFilesSubfolder,
			String filePath,
			boolean use_hash) throws DownloadException {
		
		if (u == null) {
			throw new DownloadException("Bad URL");
		}
		
		File cachedFile = null;
		
		if (type == DownloadTask.FILE) {
			
			if (use_hash) {
				cachedFile = Utils.getHashCacheFileForUrl(
						u, CACHED_FILES_ROOT_PATH, cachedFilesSubfolder, exte);
			} else {
				cachedFile = new File(CACHED_FILES_ROOT_PATH
						+ File.separator
						+ cachedFilesSubfolder
						+ File.separator
						+ filePath);
			}
			
			if (cachedFile.exists()) {
				logger.debug("CACHED FILE EXISTED !!!");
				return cachedFile; 
			}
		}

		DownloadTask task = new DownloadTask(u, exte, type, cachedFile);
		Thread down_thread = new Thread(task);
		
		down_thread.start();
		
		long time_out = MAX_TIME_ALLOWED;
		while ((time_out > 0) && (!task.isDone())) {
			try { Thread.sleep(500); } catch (InterruptedException e) {}
			time_out = time_out - MAX_TIME_ALLOWED_STEP;
		}
		
		if (task.isDone()) {
			
			Object resp = task.getDownloadedObject();
			if (resp == null) {
				task.cancel();
				throw new DownloadException("error");
			} else {
				return resp;
			}
			
		} else {
			task.cancel();
			throw new DownloadException("timeout");
		}
	}
	
	
	/**
	 * This method downloads the resource into a local File or as an array of bytes in memory.
	 * Overwrites local file if exists.
	 * 
	 * @param u the URL
	 * @param exte the extension (used if it's going to be a file)
	 * @param type download type (file/array)
	 * @param cachedFilesSubfolder folder destination (used when it's a file)
	 * @param filePath file path on the folder destination
	 * @param use_hash whether a hash function must be used to code the file name
	 * @return an object file or an array of bytes
	 * @throws DownloadException
	 */
	public static Object downloadOverwriteObjectCancel(
			URL u,
			String exte,
			int type,
			String cachedFilesSubfolder,
			String filePath,
			boolean use_hash) throws DownloadException {
		
		if (u == null) {
			throw new DownloadException("Bad URL");
		}
		
		File cachedFile = null;
		
		if (type == DownloadTask.FILE) {
			
			if (use_hash) {
				cachedFile = Utils.getHashCacheFileForUrl(
						u, CACHED_FILES_ROOT_PATH, cachedFilesSubfolder, exte);
			} else {
				cachedFile = new File(CACHED_FILES_ROOT_PATH
						+ File.separator
						+ cachedFilesSubfolder
						+ File.separator
						+ filePath);
			}
			if (cachedFile.exists()) {
				logger.debug("Cached file existed but is overwritten...");
				return cachedFile; 
			}
		}

		DownloadTask task = new DownloadTask(u, exte, type, cachedFile);
		Thread down_thread = new Thread(task);
		
		down_thread.start();
		
		long time_out = MAX_TIME_ALLOWED;
		while ((time_out > 0) && (!task.isDone())) {
			try { Thread.sleep(500); } catch (InterruptedException e) {}
			time_out = time_out - MAX_TIME_ALLOWED_STEP;
		}
		
		if (task.isDone()) {
			
			Object resp = task.getDownloadedObject();
			if (resp == null) {
				task.cancel();
				throw new DownloadException("error");
			} else {
				return resp;
			}
			
		} else {
			task.cancel();
			throw new DownloadException("timeout");
		}
	}	
	/**
	 * This method downloads a URL as a byte array. It does no participate in
	 * the file cache, so should be avoided
	 * 
	 * @param u
	 * @param exte
	 * @return the resource as a bute array
	 * @throws DownloadException
	 */
	public static byte[] getByteArray(URL u, String exte) throws DownloadException {
		
		ArrayList byteArraysList = new ArrayList(); 

		try {
			
			byte[] buffer = new byte[1024*4];
			DataInputStream is = new DataInputStream(u.openStream());
			int full_size = 0;

			for (int i = is.read(buffer); i>0; i = is.read(buffer)){
				byteArraysList.add(new Integer(i));
				byteArraysList.add(buffer.clone());
				full_size = full_size + i;
			}
			
			byteArraysList.add(new Integer(full_size));
			
		} catch (Exception e) {
			if (e.getMessage().indexOf("403") != -1) {
				throw new DownloadException("403");
			} else {
				throw new DownloadException(e.getMessage());
			}
			
		}
		
		return mergeBytes(byteArraysList);
	}

	private static byte[] mergeBytes(ArrayList byteArraysList) {
		
		int array_size = byteArraysList.size(); 
		int total_size =  ((Integer) byteArraysList.get(array_size - 1)).intValue();
		int n = array_size / 2;
		
		int item_size = 0;
		byte[] item_bytes = null;

		byte[] resp = new byte[total_size];

		int index = 0;
		for (int i=0; i<n; i++) {
			
			logger.debug("MERGE BYTES: i = " + i);
			item_size = ((Integer) byteArraysList.get(2 * i)).intValue();
			item_bytes = (byte[]) byteArraysList.get(2 * i + 1);
			System.arraycopy(item_bytes, 0, resp, index, item_size);
			index = index + item_size;
		}
		return resp;
	}
	
	
	
	public static File getResponseForPostAndRequest(
			URL url, String _req, String subfolder, boolean overw) throws Exception {
		
		StringBuffer sb = new StringBuffer();
		sb.append(url.toString());
		sb.append("_");
		sb.append(_req);
		File f = Utils.getHashCacheFileForUrlAndData(sb.toString(),
				CACHED_FILES_ROOT_PATH, subfolder, "xml");
		

		if ((!overw) && (f.exists())) {
			return f;
		}

        try {
        	HttpURLConnection conn = (HttpURLConnection) url.openConnection();
        	conn.setDoOutput(true);
        	conn.setRequestMethod("POST");
        	conn.setRequestProperty("Content-type", "application/x-www-form-urlencoded");
        	conn.setRequestProperty("Content-length", "" + _req.length());
        	OutputStreamWriter wr = new OutputStreamWriter(conn.getOutputStream());
        	wr.write(_req);
        	wr.flush();

        	DataOutputStream dos = new DataOutputStream(
        			new BufferedOutputStream(new FileOutputStream(f)));
        	byte[] buffer = new byte[1024 * 256];
        	InputStream is = conn.getInputStream();
        	long readed = 0;

        	for (int i = is.read(buffer); i > 0; i = is.read(buffer)) {
        		dos.write(buffer, 0, i);
        		readed += i;
        	}
        	dos.close();
        	is.close();
        	tryToClose(wr);
        } catch (Exception ce) {
        	logger.error("While getting req from post: " + ce.getMessage());
        	throw ce;
        }
        return f;
    }


	private static void tryToClose(OutputStreamWriter osw) {
		
		try {
			osw.close();
		} catch (Exception e) {
			logger.warn("While trying to close stream: " + e.getMessage());
		}
	}

	
}
